.S +1
.fp 5 C
.TL
IF_ENA Driver Description
.AU ""
.AF ""
.ND "June 14, 1991"
.MT 4
.SA 0
.nr Hb 4
.nr Hy 0
.nr Cl 4
.nr Hs 4
.ds HF 3 3 3 3 2 2 2
.ds HP +1 +1
.H 1 "Introduction"
The \f3if_ena\fP driver provides support for the Western Digital WD80X3 and IBMENA
families of LAN interfaces.  The driver conforms to IBM AIX PS/2
device driver requirements
for such a driver and follows the interfaces defined in the \f2AIX Operating
System For Personal System/2 Technical Reference Volume 2\fP, Appendix C.
.P
The basic driver form is based on drivers for the 4.3BSD version of UNIX.
As a result, there are few surprises to be found in the driver.
.P
This document describes the driver functions and data structures.
Data structures which are a generic part of AIX will not be described in
detail since they are documented elsewhere.
.P
The last section of this document gives a brief overview of the flow
of control through the driver.
.H 1 "Driver Functions"
There are fourteen functions used in the \f3if_ena\fP driver.  Of these functions,
most are mandated by the AIX LAN driver interface and the rest are auxilliary
functions used only by the driver.
These functions are:
\f5wdinit\fP,
\f5wdattach\fP,
\f5wdintr\fP,
\f5wdreset\fP,
\f5wdoutput\fP,
\f5wdstart\fP,
\f5wdinput\fP,
\f5wdioctl\fP,
\f5wd_insert_slot\fP,
\f5wdinitboard\fP,
\f5wdgetboardtype\fP,
\f5wdtype\fP,
\f5wdbufflen\fP,
and
\f5wdwatchdog\fP.
.P
The rest of this section will describe these functions in detail.
.H 2 "wdinit(dev)"
This function is a standard AIX driver initialization routine.  The input
parameter, \f5dev\fP, is defined for init routines and is used in
LAN interface drivers only for the major device number which is used
by the AIX \f5devexist\fP function.
.P
The init function starts by querying AIX for all boards found for each
adaptor ID defined in the array \f5wd_adaptids\fP.  As each board is found,
it is inserted into the \f5slotmap\fP array with an insertion sort on the
slot number.  After the board's presence has been recorded, a count of the number of
boards found is incremented.
.P
If no boards are found, the init function exits.
.P
If boards were found the device is installed with
.SM DEV_INSTALL.
The boards are then examined starting at the lowest slot
number found and working to the highest slot number found.  This results
in sequential unit numbers being assigned with zero for the board in the
lowest numbered slot, increasing by one for each additional board found in
slot order.
.P
As each board
is examined, the board type and configuration information is obtained and
stored in a \f5struct wdsoftc\fP structure in the array \f5wd_ifs\fP, the
driver interrupt handler is then inserted as a handler for the IRQ level
the board is configured for, the shared memory segment is mapped into
kernel address space, and the \f5wdattach\fP function is called with
the unit number corresponding to the board.
.P
When the init function exits, the driver is ready to be configured and
the hardware initialized.
.H 2 "wdattach(unit)"
This function uses its parameter, \f5unit\fP, as an index into the \f5wd_ifs\fP
array and then initializes the \f5struct ifnet\fP structure defined by
the \f5wd_if\fP member of the wdsoftc structure.
.P
It is here that the \f2name\fP and \f2unit\fP number of the device become
associated with the ifnet structure.  This is the name used by
\f2ifconfig\fP and displayed by \f2netstat -i\fP.
.H 2 "wdintr(vec)"
This function is the center of most driver activity.  It is the device
interrupt handler.  As such, it conforms to AIX device interrupt
handler requirements and supports multiple devices on the same and
different IRQ levels.
.P
When the interrupt handler is called, it searches all of its device
data structures for all devices which match the IRQ level specified by
the \f5vec\fP parameter.  For each device found, the interrupt handler
first disables the
device interrupts and then checks to see if this device did, in fact,
interrupt.  If the device didn't interrupt, the handler restores the
interrupt status to its
original value and continues the search.  This allows the IRQ to be
shared with other drivers.
.P
If the device did interrupt, a number of conditions are checked for
and the events indicated in the status word are processed.  The events
handled are:
.BL
.LI
Receive Ring Overrun
.br
In the case of a 8390 chip, special actions must be taken to let it
recover.
.LI
Packet Received
.br
The receive buffer is scanned for all received packets.  For each one
found, the \f5wdinput\fP function is called to retreive it.  Then the
packet is examined for its type and processed accordingly.
.P
If the type is a value less than 
.SM ETHERNETMTU,
it is interpreted as an
IEEE 802.2 LLC (Logical Link Control) packet.  In the current version of the driver, only SNAP
packets (802.1 Sub-Network Access Protocol) are actually interpreted.
The XID (eXchange ID) and TEST packets of the LLC protocol are
responded to if received.
.P
If the type is greater than
.SM ETHERNETMTU,
it is interpreted as an
Ethernet packet.
.P
In both data link protocol modes, 802.2 and Ethernet, only the IP
(Internet Protocol) and ARP (Address Resolution Protocol) are retained
and queued to the appropriate input queue.  ARP packets are processed
directly by calling \f5arpinput\fP and IP is placed on the \f5ipintrq\fP and
a network, software interrupt is scheduled.
.LI
Transmit Complete Interrupt
.br
When this event is received, the currently active transmit buffer is
freed and \f5wdstart\fP is called to start another packet transmitting if
possible.
.LE
.P
Statistics counters in the ifnet structure are updated appropriately
for all events.
.H 2 "wdreset(ifp)"
This function is called by \f5wdioctl\fP when the IFF_UP flag is
turned off by the \f2ifconfig\fP command.  The \f5ifp\fP parameter is used
to find the specific hardware to be reset.
.H 2 "wdoutput(ifp, m0, dst)"
This function is used to format a packet to be transmitted into the
appropriate Ethernet or IEEE format and then queue it to the hardware
specific queue.  If the hardware isn't busy, it is started by calling
the \f5wdstart\fP function.  The \f5ifp\fP parameter is used to find the
hardware specific queue in the ifnet structure.  The \f5m0\fP
parameter is an mbuf pointer pointing to the data to be transmitted.
The \f5dst\fP parameter is a poitner to an address structure which
will, in the case of an
.SM AF_INET
address, be converted to an
Ethernet/IEEE 802.3 physical address via the AIX \f5arp_resolve\fP
function.  This final address is used to build the header to the
packet.  The data is appended to the header and queued for transmit.
.H 2 "wdstart(ifp)"
This function checks for the availability of a transmit buffer.  If a
buffer is available, the packet is copied to the buffer.  If the
hardware is not busy with a previous transmit, the buffer transmit
will be started.  If there isn't an available buffer, the function
just returns
.H 2 "wdinput(ifp, wdif)
This function is called by the interrupt handler when a packet is
available to be received.  It is called once for each packet the
interrupt handler finds in the receive buffer ring.  The entire packet is
copied into an mbuf, or chain of mbufs.  The head of the resultant
mbuf chain is
returned as the result of the function.
.H 2 "wdioctl(ifp, cmd, data)
This functions is used to implement several ioctl calls specific to
LAN interfaces.
.BL
.LI
SIOCSIFADDR
.br
Set the interface address to the specified value.  Currently only the
.SM AF_INET
address family is supported.
.LI
SIOCSIFFLAGS
.br
Set the flags for the interface.  If the board is changing
.SM IFF_UP
state, the board is initiallized by calling \f5wdinitboard\fP when changing
to
.SM IFF_UP
and reset by \f5wdreset\fP when
.SM IFF_UP
is cleared.
.P
If the
.SM IF_IEEE
flag is set, the \f5if_mtu\fP member of the ifnet
structure for the device is changed from
.SM ETHERNETMTU
to
.SM ETHERMTU
to
account for the size of the IEEE 802.2 header.
.LE
.H 2 "wd_insert_slot(map, nwd, slot)"
This function is an internal utility which performs an insertion sort
to add the \f5slot\fP parameter to the list of slots which have
appropriate boards.  The purpose is to be able to have the logical
unit numbers of the interfaces correspond to the slot order of the
boards.  This is to compensate for the AIX \f5devexist\fP function
searching for boards by POS ID values.  This driver supports multiple
POS IDs.
.P
The \f5map\fP parameter is the unsigned character array of slots and
the \f5nwd\fP parameter is the size of the array.
.H 2 "wdinitboard(unit)"
This function initializes the hardware to a working state.  The
\f5unit\fP parameter identifies which board to initialize.
.P
The board is first reset to a \f2powerup\fP state and the buffer RAM is
configured to the required physical address and enabled for use.
The remaining setup consists of:
.BL
.LI
Check if a 16-bit board requiring additional setup.
.LI
If the board uses an \f2interface\fP chip, then the \f2Configuration
Control Register\fP (CCR) is initialized.
.LI
The access, receive and transmit modes are setup.
.LI
The bus transfer mode is setup.  It is always word mode.
.LI
The transmit buffer availability is configured.  The driver uses
double buffering if there is more than 8K of RAM in order to improve
performance.
.LI
The transmit and receive buffer spaces are configured along with the
memory size.
.LI
The desired interrupts are configured.
.LI
The physical address of the board is programmed into the controller
chip.
.LI
The receiver is configured and started running.
.LE
.H 2 "wdgetboardtype(port, wdif)"
This function is used to probe the board identified by the IO
\f5port\fP parameter and the \f5wdif\fP parameter.  Various tests are
done to determine what features are present. Features checked for
include:
.BL
.LI
Hardware revision level
.LI
Interface chip presence
.LI
If a 690 chip is used instead of an 8390
.LI
The board media type
.LI
If a 594 interface chip is being used
.LI
The RAM size
.LE
.H 2 "wdtype(type)"
This function uses the \f5type\fP parameter to look up the string
identifying the board type.
.H 2 "wdbufflen(m0)"
This function is a local utility routine that determines the number
of bytes in an mbuf chain.
.H 2 "wdwatchdog(wdif)"
This function is a special case to handle the 690 chip problem where
it can lock up under heavy load.  This \f2watchdog\fP timer function
is scheduled to go off HZ/2 clock ticks after the transmit is started.
If the transmit completes correctly, the timer is cancelled.
.H 1 "Driver Data Structures"
There are a few data structures specific to the driver.  The rest are
standard AIX data structures.
The important driver specific data structures are:
.BL
.LI
\f5unsigned int wd_adaptids[]\fP
.br
This array contains a list of all supported POS IDs.  It is used to
drive the search for boards via the devexist function in \f5wdinit\fP.
.P
When new boards that have different IDs become available, this array
should be added to.
.LI
\f5struct wdsoftc *wd_ifs\fP
.br
This pointer will be initialized to point to a data area which is
large enough to have one \f5wdsoftc\fP structure per board found in
the system.
.P
The wdsoftc structure contains a number of structure members which
hold the data necessary to control each board.  It also contains one
member which is the ARP table which in turn contains the standard AIX
ifnet structure for the board.  The rest of the structure members are
used to control the board
and keep track of which transmit buffers are in use.
.P
The structure is defined as follows:
.DS
.ft 5
.S -2
struct wdsoftc {
   struct arpcom wd_ac;         /* ARP table for this device */
#define wd_if   wd_ac.ac_if
#define wd_physaddr wd_ac.ac_lanaddr
   unsigned int wd_vec;         /* IRQ level */
   unsigned int wd_ioport;      /* I/O port */
   unsigned int wd_memsize;     /* size of memory segment */
   caddr_t      wd_base;        /* base address (not virtual) */
   caddr_t      wd_mp;          /* mapped address of memory segment */
   unsigned long wd_type;       /* derived hardware feature type */
   unsigned short wd_id;        /* board adaptor ID */
   unsigned short wd_rev;       /* board revision */
   struct callout *wd_watchdog; /* watchdog callout pointer */
   unsigned short wd_bufflen[2];/* length of transmit data in buffer */
   unsigned char wd_buffers;    /* number of transmit buffers */
   unsigned char wd_buffavail;  /* buffers available */
   unsigned char wd_buffinuse;  /* transmit buffers in use (mask) */
   unsigned char wd_xmitbusy;   /* tranmitter is busy */
   unsigned char wd_lastbuff;   /* last buffer used */
   unsigned char wd_nxtpkt;     /* next packet in the receive ring */
};
.S +2
.ftP
.DE
.LI
\f5struct ifnet\fP
.br
This structure is contained within the wdsoftc structure described
above.  It is contained within the \f5wd_ac\fP member of the wdsoftc
structure.  The ifnet structure is in turn linked into a list of
devices by the \f5wdattach\fP function in order for the interface to
be found by the protocols which might use the interface.
.P
The ifnet structure is the primary mechanism for associating the
device specific driver with protocol drivers.  All output gets queued
through this structure while waiting on the hardware to become ready.
Since this is a standard AIX data structure, it won't be described further.
.LE
.H 1 "Flow of Control"
There are three basic paths through which driver control flows:
intialization, output, and input.  These will be outlined separately.
.P
As was stated above, the ifnet structure is the common tie between
protocols and the device driver.
.H 2 "Initialization Path"
Initialization starts at system boot time and is completed when the
interface is configured.
.P
Essentially, \f5wdinit\fP is called at boot time which uses some standard
AIX utilities and then determines how many boards and of which types
are present in the system.
.P
At configuration time, \f5wdioctl\fP is called which will then call
\f5wdinitboard\fP to configure and activate the specific hardware.
.H 2 "Output Path"
The users of the interface are protocol stacks.  When a protocol wants
to send data, it calls the \f5wdoutput\fP function which does appropriate
manipulation of the data, queues it and then tries to send it if the
hardware is available.  If it isn't available at that time, \f5wdoutput\fP
returns.  In this case, another attempt will be made when the
transmitter hardware interrupts on transmit complete and \f5wdstart\fP is
called to try sending the queued data.
.H 2 "Input Path"
This path is activated on a \f2receive\fP interrupt.  The \f5wdintr\fP
function is called to handle the interrupt.  This can result in
\f5wdinput\fP being called to retrieve the packet.  If the packet is one the
system is interested in, either \f5arpinput\fP will be called or it will be
inserted into the protocol receive queue and a software interrupt for
that protocol will be scheduled.
